from typing import Dict
import requests as r
from urllib.parse import urlparse
import os.path
import re
import sys

from .data.files_map import files_map, inverse_files_map
from .data.sites import sites


def extract_etag(url: str, etags: Dict[str, str]):
    res = r.get(
        url,
        stream=True
    )
    if res.status_code != 200:
        raise RuntimeError(f"Panic: failed to get {url}: {res.status_code}")

    etag = res.headers["ETag"]
    res.close()

    parsed_url = urlparse(url)
    path = parsed_url.path
    filename = os.path.basename(path)

    etags[filename] = etag

    print(f"ETag for {filename}: {etag}")


def get_file_name(site_or_url: str) -> str:
    domain = re.sub(r'https://', '', site_or_url)

    try:
        file_name = files_map[domain]
        return f'{file_name}.7z'
    except KeyError:
        return f'{domain}.7z'


def is_dump_file(file_name: str) -> bool:
    file_name = re.sub(r'\.7z$', '', file_name)

    try:
        inverse_files_map[file_name]
    except KeyError:
        origin = f'https://{file_name}'
        return origin in sites

    return True


def check_file(base_path: str, file_name: str) -> bool:
    try:
        res = os.stat(os.path.join(base_path, file_name))
        return res.st_size > 0
    except FileNotFoundError:
        return False


def archive_file(base_path: str, site_or_url: str) -> None:
    try:
        file_name = get_file_name(site_or_url)
        file_path = os.path.join(base_path, file_name)
        os.rename(file_path, f"{file_path}.old")
    except FileNotFoundError:
        pass


def cleanup_archive(base_path: str) -> None:
    try:
        file_entries = os.listdir(base_path)

        for entry in file_entries:
            if entry.endswith('.old'):
                entry_path = os.path.join(base_path, entry)
                os.remove(entry_path)
    except:
        print(sys.exc_info())


def is_file_downloaded(base_path: str, site_or_url: str) -> bool:
    file_name = get_file_name(site_or_url)
    return check_file(base_path, file_name)



# for most sites, dump names correspond to domain name.
# However, due to some subdomain shenanigans, a couple of sites differ:
files_map: dict[str, str] = {
    'alcohol.stackexchange.com': 'beer.stackexchange.com',
    'alcohol.meta.stackexchange.com': 'beer.meta.stackexchange.com',
    'mattermodeling.stackexchange.com': 'materials.stackexchange.com',
    'mattermodeling.meta.stackexchange.com': 'materials.meta.stackexchange.com',
    'communitybuilding.stackexchange.com': 'moderators.stackexchange.com',
    'communitybuilding.meta.stackexchange.com': 'moderators.meta.stackexchange.com',
    'medicalsciences.stackexchange.com': 'health.stackexchange.com',
    'medicalsciences.meta.stackexchange.com': 'health.meta.stackexchange.com',
    'psychology.stackexchange.com': 'cogsci.stackexchange.com',
    'psychology.meta.stackexchange.com': 'cogsci.meta.stackexchange.com',
    'writing.stackexchange.com': 'writers.stackexchange.com',
    'writing.meta.stackexchange.com': 'writers.meta.stackexchange.com',
    'video.stackexchange.com': 'avp.stackexchange.com',
    'video.meta.stackexchange.com': 'avp.meta.stackexchange.com',
    'meta.es.stackoverflow.com': 'es.meta.stackoverflow.com',
    'meta.ja.stackoverflow.com': 'ja.meta.stackoverflow.com',
    'meta.pt.stackoverflow.com': 'pt.meta.stackoverflow.com',
    'meta.ru.stackoverflow.com': 'ru.meta.stackoverflow.com',
}

inverse_files_map: dict[str, str] = {v: k for k, v in files_map.items()}



from selenium.webdriver import Firefox
from sys import exc_info
from time import sleep

from ..config import SEDDConfig
from .utils import ubo_set_user_settings, \
    ubo_set_advanced_settings, ubo_set_selected_filters, \
    ubo_set_whitelist, ubo_set_dynamic_rules, ubo_set_user_filters, \
    ubo_reload_all_filters


def init_ubo_settings(browser: Firefox, config: SEDDConfig, ubo_id: str) -> bool:
    try:
        browser.get(
            f'moz-extension://{ubo_id}/dashboard.html#settings.html'
        )

        settings = config.get_ubo_settings()

        ubo_set_user_settings(browser, settings)
        ubo_set_advanced_settings(browser, settings)

        # idk why, but applyFilterListSelection only works after a delay
        sleep(1)

        ubo_set_selected_filters(browser, settings)
        ubo_set_whitelist(browser, settings)
        ubo_set_dynamic_rules(browser, settings)
        ubo_set_user_filters(browser, settings)

        ubo_reload_all_filters(browser)
    except:
        print('Failed to set uBLock config, using defaults')
        print(exc_info())



from os import path, makedirs
from urllib import request
from json import dumps
from uuid import uuid4
import platform

from selenium import webdriver
from selenium.webdriver.firefox.options import Options
from undetected_geckodriver import Firefox as UFirefox

from .config import SEDDConfig
from .ubo import init_ubo_settings


def init_output_dir(output_dir: str):
    if not path.exists(output_dir):
        makedirs(output_dir)

    print(output_dir)

    return output_dir


def init_firefox_driver(config: SEDDConfig, disable_undetected: bool, output_dir: str):
    options = Options()
    options.enable_downloads = True
    options.set_preference("browser.download.folderList", 2)
    options.set_preference("browser.download.manager.showWhenStarting", False)
    options.set_preference("browser.download.dir", output_dir)
    options.set_preference(
        "browser.helperApps.neverAsk.saveToDisk", "application/x-gzip"
    )

    # our own uuid for uBO so as we don't need to do the dance of inspecing internals
    ubo_internal_uuid = f"{uuid4()}"

    options.set_preference("extensions.webextensions.uuids", dumps(
        {"uBlock0@raymondhill.net": ubo_internal_uuid}))

    is_apple = platform.system() == "Darwin"
    use_undetected = not disable_undetected and not is_apple
    if use_undetected:
        print("Using undetected-geckodriver")
        browser = UFirefox(options = options)
    else:
        print("Warning: using standard geckodriver. Cloudflare may perpetually block you")
        if is_apple:
            print("This option is forced on macOS. For undetected_geckodriver, "
                  "run the downloader in a Linux or Windows environment.")
        browser = webdriver.Firefox(options=options)

    ubo_download_url = config.get_ubo_download_url()

    if not path.exists("ubo.xpi"):
        print(f"Downloading uBO from: {ubo_download_url}")
        request.urlretrieve(ubo_download_url, "ubo.xpi")

    ubo_id = browser.install_addon("ubo.xpi", temporary=True)

    init_ubo_settings(browser, config, ubo_internal_uuid)

    return browser, ubo_id



import argparse

from os import getcwd
from os.path import join


class SEDDCLIArgs(argparse.Namespace):
    skip_loaded: bool
    keep_consent: bool
    output_dir: str
    dry_run: bool
    disable_undetected: bool
    verbose: bool


parser = argparse.ArgumentParser(
    prog="sedd",
    description="Automatic (unofficial) SE data dump downloader for the anti-community data dump format",
)

parser.add_argument(
    "-s", "--skip-loaded",
    required=False,
    default=False,
    action="store_true",
    dest="skip_loaded"
)

parser.add_argument(
    "-k", "--keep-consent",
    required=False,
    dest="keep_consent",
    action="store_true",
    default=False
)

parser.add_argument(
    "-o", "--outputDir",
    required=False,
    dest="output_dir",
    default=join(getcwd(), "downloads")
)

parser.add_argument(
    "-g", "--disable-undetected-geckodriver",
    required=False,
    dest="disable_undetected",
    action="store_true",
    default=False
)

parser.add_argument(
    "--dry-run",
    required=False,
    default=False,
    action="store_true",
    dest="dry_run"
)
parser.add_argument(
    "-v",
    required=False,
    default=False,
    action="store_true",
    dest="verbose"
)


def parse_cli_args() -> SEDDCLIArgs:
    return parser.parse_args()



from threading import current_thread, main_thread
from watchdog.observers import Observer

from .handler import CleanupHandler
from .state import DownloadState


def register_pending_downloads_observer(output_dir: str):
    if current_thread() is main_thread():
        observer = Observer()
        state = DownloadState()
        handler = CleanupHandler(observer, state)

        observer.schedule(handler, output_dir, recursive=True)
        observer.start()

        return state, observer



from desktop_notifier import DesktopNotifier
import asyncio

from ..config import SEDDConfig


def native(message: str, _):
    asyncio.run(
        DesktopNotifier().send(
            title="The data dump downloader needs attention",
            message=f"{message}"
        )
    )


notification_providers = {
    "native": native
}


def notify(message: str, config: SEDDConfig):

    provider = config.get_notifications_provider()

    if provider is None:
        print(message)
        return

    notification_providers[provider](message, config.notifications)



from .typings import SEDDNotificationsConfig, SEDDUboSettings, SEDDUboConfig

default_ubo_url = "https://github.com/gorhill/uBlock/releases/download/1.59.0/uBlock0_1.59.0.firefox.signed.xpi"

default_notifications_config: SEDDNotificationsConfig = {
    'provider': None
}

default_ubo_settings: SEDDUboSettings = {
    'userSettings': {
        'advancedUserEnabled': True,
    },
    "hiddenSettings": {},
    "selectedFilterLists": [
        "user-filters",
        "ublock-filters",
        "ublock-badware",
        "ublock-privacy",
        "ublock-unbreak",
        "ublock-quick-fixes",
        "easylist",
        "easyprivacy",
        "urlhaus-1",
        "plowe-0"
    ],
    "whitelist": [
        "chrome-extension-scheme",
        "moz-extension-scheme"
    ],
    "dynamicFilters": {
        "toAdd": [],
        "toRemove": []
    },
    "userFilters": {
        "enabled": False,
        "trusted": False,
        "toOverwrite": []
    }
}

default_ubo_config: SEDDUboConfig = {
    'download_url': default_ubo_url,
    'settings': default_ubo_settings
}



# Beta sites don't have data dumps. They're included in this comment for the record:
# https://cs50.stackexchange.com
sites = [
    "https://stackoverflow.com",
    "https://3dprinting.stackexchange.com",
    "https://academia.stackexchange.com",
    "https://ai.stackexchange.com",
    "https://alcohol.stackexchange.com",
    "https://android.stackexchange.com",
    "https://anime.stackexchange.com",
    "https://apple.stackexchange.com",
    "https://arduino.stackexchange.com",
    "https://askubuntu.com",
    "https://astronomy.stackexchange.com",
    "https://aviation.stackexchange.com",
    "https://bicycles.stackexchange.com",
    "https://bioacoustics.stackexchange.com",
    "https://bioinformatics.stackexchange.com",
    "https://biology.stackexchange.com",
    "https://bitcoin.stackexchange.com",
    "https://blender.stackexchange.com",
    "https://boardgames.stackexchange.com",
    "https://bricks.stackexchange.com",
    "https://buddhism.stackexchange.com",
    "https://cardano.stackexchange.com",
    "https://chemistry.stackexchange.com",
    "https://chess.stackexchange.com",
    "https://chinese.stackexchange.com",
    "https://christianity.stackexchange.com",
    "https://civicrm.stackexchange.com",
    "https://codegolf.stackexchange.com",
    "https://codereview.stackexchange.com",
    "https://coffee.stackexchange.com",
    "https://communitybuilding.stackexchange.com",
    "https://computergraphics.stackexchange.com",
    "https://conlang.stackexchange.com",
    "https://cooking.stackexchange.com",
    "https://craftcms.stackexchange.com",
    "https://crafts.stackexchange.com",
    "https://crypto.stackexchange.com",
    "https://cs.stackexchange.com",
    "https://cseducators.stackexchange.com",
    "https://cstheory.stackexchange.com",
    "https://datascience.stackexchange.com",
    "https://dba.stackexchange.com",
    "https://devops.stackexchange.com",
    "https://diy.stackexchange.com",
    "https://drones.stackexchange.com",
    "https://drupal.stackexchange.com",
    "https://dsp.stackexchange.com",
    "https://earthscience.stackexchange.com",
    "https://ebooks.stackexchange.com",
    "https://economics.stackexchange.com",
    "https://electronics.stackexchange.com",
    "https://elementaryos.stackexchange.com",
    "https://ell.stackexchange.com",
    "https://emacs.stackexchange.com",
    "https://engineering.stackexchange.com",
    "https://english.stackexchange.com",
    "https://eosio.stackexchange.com",
    "https://es.stackoverflow.com",
    "https://esperanto.stackexchange.com",
    "https://ethereum.stackexchange.com",
    "https://expatriates.stackexchange.com",
    "https://expressionengine.stackexchange.com",
    "https://fitness.stackexchange.com",
    "https://freelancing.stackexchange.com",
    "https://french.stackexchange.com",
    "https://gamedev.stackexchange.com",
    "https://gaming.stackexchange.com",
    "https://gardening.stackexchange.com",
    "https://genai.stackexchange.com",
    "https://genealogy.stackexchange.com",
    "https://german.stackexchange.com",
    "https://gis.stackexchange.com",
    "https://graphicdesign.stackexchange.com",
    "https://ham.stackexchange.com",
    "https://hardwarerecs.stackexchange.com",
    "https://hermeneutics.stackexchange.com",
    "https://hinduism.stackexchange.com",
    "https://history.stackexchange.com",
    "https://homebrew.stackexchange.com",
    "https://hsm.stackexchange.com",
    "https://interpersonal.stackexchange.com",
    "https://iot.stackexchange.com",
    "https://iota.stackexchange.com",
    "https://islam.stackexchange.com",
    "https://italian.stackexchange.com",
    "https://ja.stackoverflow.com",
    "https://japanese.stackexchange.com",
    "https://joomla.stackexchange.com",
    "https://judaism.stackexchange.com",
    "https://korean.stackexchange.com",
    "https://langdev.stackexchange.com",
    "https://languagelearning.stackexchange.com",
    "https://latin.stackexchange.com",
    "https://law.stackexchange.com",
    "https://lifehacks.stackexchange.com",
    "https://linguistics.stackexchange.com",
    "https://literature.stackexchange.com",
    "https://magento.stackexchange.com",
    "https://martialarts.stackexchange.com",
    "https://math.stackexchange.com",
    "https://matheducators.stackexchange.com",
    "https://mathematica.stackexchange.com",
    "https://mathoverflow.net",
    "https://mattermodeling.stackexchange.com",
    "https://mechanics.stackexchange.com",
    "https://medicalsciences.stackexchange.com",
    "https://meta.stackexchange.com",
    "https://monero.stackexchange.com",
    "https://money.stackexchange.com",
    "https://movies.stackexchange.com",
    "https://music.stackexchange.com",
    "https://musicfans.stackexchange.com",
    "https://mythology.stackexchange.com",
    "https://networkengineering.stackexchange.com",
    "https://opendata.stackexchange.com",
    "https://opensource.stackexchange.com",
    "https://or.stackexchange.com",
    "https://outdoors.stackexchange.com",
    "https://parenting.stackexchange.com",
    "https://patents.stackexchange.com",
    "https://pets.stackexchange.com",
    "https://philosophy.stackexchange.com",
    "https://photo.stackexchange.com",
    "https://physics.stackexchange.com",
    "https://pm.stackexchange.com",
    "https://poker.stackexchange.com",
    "https://politics.stackexchange.com",
    "https://portuguese.stackexchange.com",
    "https://proofassistants.stackexchange.com",
    "https://psychology.stackexchange.com",
    "https://pt.stackoverflow.com",
    "https://puzzling.stackexchange.com",
    "https://quant.stackexchange.com",
    "https://quantumcomputing.stackexchange.com",
    "https://raspberrypi.stackexchange.com",
    "https://retrocomputing.stackexchange.com",
    "https://reverseengineering.stackexchange.com",
    "https://robotics.stackexchange.com",
    "https://rpg.stackexchange.com",
    "https://ru.stackoverflow.com",
    "https://rus.stackexchange.com",
    "https://russian.stackexchange.com",
    "https://salesforce.stackexchange.com",
    "https://scicomp.stackexchange.com",
    "https://scifi.stackexchange.com",
    "https://security.stackexchange.com",
    "https://serverfault.com",
    "https://sharepoint.stackexchange.com",
    "https://sitecore.stackexchange.com",
    "https://skeptics.stackexchange.com",
    "https://softwareengineering.stackexchange.com",
    "https://softwarerecs.stackexchange.com",
    "https://solana.stackexchange.com",
    "https://sound.stackexchange.com",
    "https://space.stackexchange.com",
    "https://spanish.stackexchange.com",
    "https://sports.stackexchange.com",
    "https://sqa.stackexchange.com",
    "https://stackapps.com",
    "https://stats.stackexchange.com",
    "https://stellar.stackexchange.com",
    "https://substrate.stackexchange.com",
    "https://superuser.com",
    "https://sustainability.stackexchange.com",
    "https://tex.stackexchange.com",
    "https://tezos.stackexchange.com",
    "https://tor.stackexchange.com",
    "https://travel.stackexchange.com",
    "https://tridion.stackexchange.com",
    "https://ukrainian.stackexchange.com",
    "https://unix.stackexchange.com",
    "https://ux.stackexchange.com",
    "https://vegetarianism.stackexchange.com",
    "https://vi.stackexchange.com",
    "https://video.stackexchange.com",
    "https://webapps.stackexchange.com",
    "https://webmasters.stackexchange.com",
    "https://woodworking.stackexchange.com",
    "https://wordpress.stackexchange.com",
    "https://workplace.stackexchange.com",
    "https://worldbuilding.stackexchange.com",
    "https://writing.stackexchange.com",
]

# For testing
# sites = ["https://stackoverflow.com"]



from typing import TypedDict, Literal


class SEDDNotificationsConfig(TypedDict):
    provider: Literal['native'] | None


class SEDDUboUserFiltersSettings(TypedDict):
    enabled: bool
    trusted: bool
    toOverwrite: list[str]


class SEDDUboSettings(TypedDict):
    userSettings: dict[str, str | bool | int]
    hiddenSettings: dict[str, str | bool | int]
    selectedFilterLists: list[str]
    whitelist: list[str]
    dynamicFilters: dict[Literal['toAdd'] | Literal['toRemove'], list[str]]
    userFilters: SEDDUboUserFiltersSettings


class SEDDUboConfig(TypedDict):
    download_url: str
    settings: SEDDUboSettings | None



from typing import Literal

from json import load
from os import path, getcwd

from .defaults import default_ubo_url, default_ubo_settings, default_notifications_config, default_ubo_config
from .typings import SEDDNotificationsConfig, SEDDUboConfig, SEDDUboSettings


class SEDDConfig:
    email: str
    password: str
    notifications: SEDDNotificationsConfig
    ubo: SEDDUboConfig

    def __init__(self, email: str, pwd: str, notifications: SEDDNotificationsConfig, ubo: SEDDUboConfig):
        self.email = email
        self.password = pwd
        self.notifications = notifications
        self.ubo = ubo

    def get_notifications_provider(self) -> Literal['native'] | None:
        notifications_config = self.notifications
        return notifications_config['provider'] if 'provider' in notifications_config else None

    def get_ubo_download_url(self) -> str:
        ubo_config = self.ubo
        return ubo_config['download_url'] if 'download_url' in ubo_config else default_ubo_url

    def get_ubo_settings(self) -> SEDDUboSettings:
        ubo_config = self.ubo
        return ubo_config['settings'] if 'settings' in ubo_config else default_ubo_settings


def load_sedd_config() -> SEDDConfig:
    config_path = path.join(getcwd(), 'config.json')

    if not path.exists(config_path):
        print("It appears you haven't made ./config.json")
        print("For a template, see:")
        print("\thttps://github.com/LunarWatcher/se-data-dump-transformer/blob/master/config.example.json")
        print("Please configure it, then try again. For further information, "
              "see the README in the Git repository.")

        exit(-1)

    with open(config_path, "r") as f:
        config = load(f)

        email = config["email"]
        password = config["password"]

        notifications = config['notifications'] if 'notifications' in config else default_notifications_config

        ubo = config['ubo'] if 'ubo' in config else default_ubo_config

        config = SEDDConfig(email, password, notifications, ubo)

    return config



from selenium.webdriver import Firefox
from ..config.typings import SEDDUboSettings


def ubo_set_user_settings(browser: Firefox, settings: SEDDUboSettings) -> None:
    # using vAPI 'userSettings' message to set uBO's user settings
    # see: https://github.com/gorhill/uBlock/blob/master/src/js/settings.js#L215
    for key, val in settings['userSettings'].items():
        browser.execute_script("""
            const name = arguments[0];
            const value = arguments[1];

            globalThis.vAPI.messaging.send('dashboard', {
                what: 'userSettings', name, value,
            });
        """, key, val)


def ubo_set_advanced_settings(browser: Firefox, settings: SEDDUboSettings) -> None:
    # using vAPI 'writeHiddenSettings' message to set uBO's advanced settings
    # see: https://github.com/gorhill/uBlock/blob/master/src/js/advanced-settings.js#L177
    browser.execute_script("""
        const settings = arguments[0];

        const content = Object.entries(settings)
                           .map(([k,v]) => `${k} ${v}`)
                           .join('\\n');

        globalThis.vAPI.messaging.send('dashboard', {
            what: 'writeHiddenSettings', content,
        });
    """, settings['hiddenSettings'])


def ubo_set_selected_filters(browser: Firefox, settings: SEDDUboSettings) -> None:
    # using vAPI 'applyFilterListSelection' message to set uBO's selected filter lists
    # see: https://github.com/gorhill/uBlock/blob/master/src/js/storage.js#L486
    browser.execute_script("""
        const toSelect = arguments[0];

        globalThis.vAPI.messaging.send('dashboard', {
            what: 'applyFilterListSelection', toSelect,
        })
    """, settings['selectedFilterLists'])


def ubo_set_whitelist(browser: Firefox, settings: SEDDUboSettings) -> None:
    # using vAPI 'setWhitelist' message to set uBO's trusted sites list
    # see: https://github.com/gorhill/uBlock/blob/master/src/js/messaging.js#L225
    browser.execute_script("""
        const list = arguments[0];

        globalThis.vAPI.messaging.send('dashboard', {
            what: 'setWhitelist', whitelist: list.join('\\n')
        })
    """, settings["whitelist"])


def ubo_set_dynamic_rules(browser: Firefox, settings: SEDDUboSettings) -> None:
    # using vAPI 'modifyRuleset' message to set uBO's dynamic rules
    # see: https://github.com/gorhill/uBlock/blob/master/src/js/dyna-rules.js#L279
    browser.execute_script("""
        const { toAdd = [], toRemove = [] } = arguments[0]

        globalThis.vAPI.messaging.send('dashboard', {
            what: 'modifyRuleset', permanent: true,
            toAdd: toAdd.join('\\n'),
            toRemove: toRemove.join('\\n'),
        })

    """, settings["dynamicFilters"])


def ubo_set_user_filters(browser: Firefox, settings: SEDDUboSettings) -> None:
    # using vAPI 'writeUserFilters' message to set uBO's user filters
    # see: https://github.com/gorhill/uBlock/blob/master/src/js/storage.js#L582
    browser.execute_script("""
        const { trusted, enabled, toOverwrite = [] } = arguments[0]

        globalThis.vAPI.messaging.send('dashboard', {
            what: 'writeUserFilters', trusted, enabled,
            content: toOverwrite.join('\\n')
        })
    """, settings['userFilters'])


def ubo_reload_all_filters(browser: Firefox) -> None:
    browser.execute_async_script("""
        const done = arguments[0]

        globalThis.vAPI.messaging.send('dashboard', {
            what: 'reloadAllFilters',
        }).then(done)
    """)



from typing import Set


class DownloadState:
    # list of filenames pending download
    pending: Set[str] = set()

    def size(self):
        return len(self.pending)

    def empty(self):
        return self.size() == 0

    def add(self, file: str):
        self.pending.add(file)

    def remove(self, file: str):
        self.pending.remove(file)


download_state = DownloadState()



from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.webdriver import WebDriver
from selenium.common.exceptions import NoSuchElementException
from typing import Dict


from time import sleep

import re
import sys
from traceback import print_exception


from .cli import parse_cli_args
from .config import load_sedd_config
from .data import sites
from .meta import notifications
from .watcher.observer import register_pending_downloads_observer
from . import utils

from .driver import init_output_dir, init_firefox_driver
import logging

args = parse_cli_args()

if args.verbose:
    logging.basicConfig(level=logging.DEBUG)

sedd_config = load_sedd_config()

output_dir = init_output_dir(args.output_dir)

browser, ubo_id = init_firefox_driver(
    sedd_config,
    args.disable_undetected,
    output_dir
)


def kill_cookie_shit(browser: WebDriver):
    sleep(3)
    browser.execute_script(
        """let elem = document.getElementById("onetrust-banner-sdk"); if (elem) { elem.parentNode.removeChild(elem); }""")
    sleep(1)

def check_cloudflare_intercept(browser: WebDriver):
    if browser.title == "Just a moment...":
        print("CF verification hit. Trying soft workaround")
        sleep(15)

        if (browser.title == "Just a moment..."):
            print("Irrecoverable state suspected; captcha solving likely required")
            notifications.notify("CloudFlare verification hit; auto-verification failed. Please complete the captcha", sedd_config)
        else:
            print("Auto-recovered from CF wall")
            return

        while browser.title == "Just a moment...":
            print("Still stuck on CF verification. Waiting for 10 seconds")
            sleep(10)

def is_logged_in(browser: WebDriver, site: str):
    url = f"{site}/users/current"
    browser.get(url)
    sleep(1)
    check_cloudflare_intercept(browser)

    return "/users/" in browser.current_url


def login_or_create(browser: WebDriver, site: str):
    if is_logged_in(browser, site):
        print("Already logged in")
    else:
        print("Not logged in and/or not registered. Logging in now")
        while True:
            browser.get(f"{site}/users/login")
            check_cloudflare_intercept(browser)

            if "?newreg" in browser.current_url:
                print(f"Auto-created {site} without login needed")
                break

            email_elem = browser.find_element(By.ID, "email")
            password_elem = browser.find_element(By.ID, "password")
            email_elem.send_keys(sedd_config.email)
            password_elem.send_keys(sedd_config.password)
            retryLogin = False

            curr_url = browser.current_url
            browser.find_element(By.ID, "submit-button").click()

            try:
                elem = browser.find_element(By.CSS_SELECTOR, "#login-form > .js-error-message")
                if elem is not None:
                    print("Login failed quietly. Retrying")
                    continue
            except:
                # No error element
                pass

            check_cloudflare_intercept(browser)

            while browser.current_url == curr_url:
                sleep(3)


            captcha_walled = False
            while "/nocaptcha" in browser.current_url:
                if not captcha_walled:
                    captcha_walled = True

                notifications.notify(
                    "Captcha wall hit during login", sedd_config
                )

                sleep(10)

            if captcha_walled or retryLogin:
                continue

            if not is_logged_in(browser, site):
                raise RuntimeError("Login failed")

            break


def download_data_dump(browser: WebDriver, site: str, meta_url: str, etags: Dict[str, str]):
    print(f"Downloading data dump from {site}")

    def _exec_download(browser: WebDriver):
        if args.keep_consent:
            print('Consent dialog will not be auto-removed')
        else:
            kill_cookie_shit(browser)

        try:
            checkbox = browser.find_element(By.ID, "datadump-agree-checkbox")
            btn = browser.find_element(By.ID, "datadump-download-button")
        except NoSuchElementException:
            raise RuntimeError(f"Bad site: {site}")

        if args.dry_run:
            return

        browser.execute_script("""
        (function() {
            let oldFetch = window.fetch;
            window.fetch = (url, opts) => {
                let promise = oldFetch(url, opts);

                if (url.includes("/link")) {
                    promise.then(res => {
                        res.clone().json().then(json => {
                            window.extractedUrl = json["url"];
                            console.log(extractedUrl);
                        });
                        return res;
                    });
                    return new Promise(resolve => setTimeout(resolve, 4000))
                        .then(_ => promise);
                }
                return promise;
            };
        })();
        """)

        checkbox.click()
        sleep(1)
        btn.click()
        check_cloudflare_intercept(browser)
        sleep(2)
        url = browser.execute_script("return window.extractedUrl;")
        utils.extract_etag(url, etags)

        sleep(5)

    main_loaded = utils.is_file_downloaded(args.output_dir, site)
    meta_loaded = utils.is_file_downloaded(args.output_dir, meta_url)

    if not args.skip_loaded or not main_loaded or not meta_loaded:
        if args.skip_loaded and main_loaded:
            pass
        else:
            browser.get(f"{site}/users/data-dump-access/current")
            check_cloudflare_intercept(browser)

            if not args.dry_run:
                utils.archive_file(args.output_dir, site)

            _exec_download(browser)

        if args.skip_loaded and meta_loaded:
            pass
        else:
            browser.get(f"{meta_url}/users/data-dump-access/current")
            check_cloudflare_intercept(browser)

            if not args.dry_run:
                utils.archive_file(args.output_dir, meta_url)

            _exec_download(browser)


etags: Dict[str, str] = {}

try:
    state, observer = register_pending_downloads_observer(args.output_dir)

    for site in sites.sites:
        if site not in ["https://meta.stackexchange.com", "https://stackapps.com"]:
            # https://regex101.com/r/kG6nTN/1
            meta_url = re.sub(
                r"(https://(?:[^.]+\.(?=stackexchange))?)", r"\1meta.", site)

        main_loaded = utils.is_file_downloaded(args.output_dir, site)
        meta_loaded = utils.is_file_downloaded(args.output_dir, meta_url)

        if args.skip_loaded and main_loaded and meta_loaded:
            pass
        else:
            print(f"Extracting from {site}...")

            login_or_create(browser, site)
            download_data_dump(
                browser,
                site,
                meta_url,
                etags
            )

    if observer:
        pending = state.size()

        print(f"Waiting for {pending} download{'s'[:pending^1]} to complete")

        while True:
            if state.empty():
                observer.stop()
                browser.quit()

                utils.cleanup_archive(args.output_dir)
                break
            else:
                sleep(1)

except KeyboardInterrupt:
    pass

except:
    exception = sys.exc_info()

    try:
        print_exception(exception)
    except:
        print(exception)

    browser.quit()
finally:
    # TODO: replace with validation once downloading is verified done
    # (or export for separate, later verification)
    # Though keeping it here, removing files and re-running downloads feels like a better idea
    print(etags)



import os

from watchdog.observers.api import BaseObserverSubclassCallable
from watchdog.events import FileSystemEventHandler

from .state import DownloadState
from ..utils import is_dump_file


class CleanupHandler(FileSystemEventHandler):
    download_state: DownloadState
    observer: BaseObserverSubclassCallable

    def __init__(self, observer: BaseObserverSubclassCallable, state: DownloadState):
        super()

        self.download_state = state
        self.observer = observer

    def on_created(self, event):
        file_name = os.path.basename(event.src_path)

        # # we can safely ignore part file creations
        if file_name.endswith('.part'):
            return

        if is_dump_file(file_name):
            print(f"Download started: {file_name}")
            self.download_state.add(file_name)

    def on_moved(self, event):
        file_name: str = os.path.basename(event.dest_path)

        # we can safely ignore part file removals
        if file_name.endswith('.part'):
            return

        if is_dump_file(file_name):
            print(f"Download finished: {file_name}")
            self.download_state.remove(file_name)
